@@ -8,6 +8,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
+#include <errno.h>
 #include <fcntl.h>
 #include <sys/types.h>
 #include <sys/socket.h>
@@ -25,6 +26,8 @@
 int closesocket(int);
 #endif
 
+#include "beos.h"
+
 int be_read(int s, void *ptr, int len)
 {
 	if (s >= SHS) return recv(s - SHS, ptr, len, 0);
@@ -80,10 +83,10 @@ int be_accept(int s, struct sockaddr *sa, int *sal)
 int be_bind(int s, struct sockaddr *sa, int sal)
 {
 	/*struct sockaddr_in *sin = (struct sockaddr_in *)sa;
-	if (!sin->sin_port) {
+	if (!ntohs(sin->sin_port)) {
 		int i;
 		for (i = 16384; i < 49152; i++) {
-			sin->sin_port = i;
+			sin->sin_port = htons(i);
 			if (!be_bind(s, sa, sal)) return 0;
 		}
 		return -1;
@@ -112,7 +115,7 @@ int be_pipe(int *fd)
 	}
 	memset(&sa1, 0, sizeof(sa1));
 	sa1.sin_family = AF_INET;
-	sa1.sin_port = 0;
+	sa1.sin_port = htons(0);
 	sa1.sin_addr.s_addr = INADDR_ANY;
 	if (be_bind(s1, (struct sockaddr *)&sa1, sizeof(sa1))) {
 		/*perror("bind");*/
@@ -154,6 +157,7 @@ int be_select(int n, struct fd_set *rd, struct fd_set *wr, struct fd_set *exc, s
 {
 	int i, s;
 	struct fd_set d, rrd;
+retry:
 	FD_ZERO(&d);
 	if (!rd) rd = &d;
 	if (!wr) wr = &d;
@@ -167,6 +171,7 @@ int be_select(int n, struct fd_set *rd, struct fd_set *wr, struct fd_set *exc, s
 	FD_ZERO(&rrd);
 	for (i = SHS; i < n; i++) if (FD_ISSET(i, rd)) FD_SET(i - SHS, &rrd);
 	if ((s = select(FD_SETSIZE, &rrd, &d, &d, tm)) < 0) {
+		if (errno == EINTR) goto retry;
 		FD_ZERO(rd);
 		return 0;
 	}
@@ -189,15 +194,11 @@ int be_getsockopt(int s, int level, int optname, void *optval, int *optlen)
 	return -1;
 }
 
-int start_thr(void (*)(void *), void *, unsigned char *);
+static int ihpipe[2];
 
-int ihpipe[2];
+static int inth;
 
-int inth;
-
-#include <errno.h>
-
-void input_handle_th(void *p)
+static void input_handle_th(void *p)
 {
 	char c;
 	int b = 0;
@@ -205,7 +206,7 @@ void input_handle_th(void *p)
 	while (1) if (read(0, &c, 1) == 1) be_write(ihpipe[1], &c, 1);
 }
 
-int get_input_handle()
+int get_input_handle(void)
 {
 	static int h = -1;
 	if (h >= 0) return h;
@@ -213,17 +214,19 @@ int get_input_handle()
 	if ((inth = start_thr(input_handle_th, NULL, "input_thread")) < 0) {
 		closesocket(ihpipe[0]);
 		closesocket(ihpipe[1]);
-		return -1;
+		error("Can't spawn input thread");
+		fatal_tty_exit();
+		exit(4);
 	}
 	return h = ihpipe[0];
 }
 
-void block_stdin()
+void block_stdin(void)
 {
 	suspend_thread(inth);
 }
 
-void unblock_stdin()
+void unblock_stdin(void)
 {
 	resume_thread(inth);
 }
@@ -232,7 +235,7 @@ void unblock_stdin()
 
 #define O_BUF	16384
 
-void output_handle_th(void *p)
+static void output_handle_th(void *p)
 {
 	char *c = malloc(O_BUF);
 	int r, b = 0;
@@ -242,7 +245,7 @@ void output_handle_th(void *p)
 	free(c);
 }
 
-int get_output_handle()
+int get_output_handle(void)
 {
 	static int h = -1;
 	if (h >= 0) return h;
@@ -250,9 +253,15 @@ int get_output_handle()
 	if (start_thr(output_handle_th, NULL, "output_thread") < 0) {
 		closesocket(ohpipe[0]);
 		closesocket(ohpipe[1]);
-		return -1;
+		error("Can't spawn output thread");
+		fatal_tty_exit();
+		exit(4);
 	}
 	return h = ohpipe[1];
 }*/
 
+#else
+
+typedef int beos_c_no_empty_unit;
+
 #endif
