@@ -12,28 +12,9 @@
 
 #include "links.h"
 
-/* prototypes */
-int gray (int, int, int);
-int too_near(int, int, int, int, int, int);
-void separate_fg_bg(int *, int *, int *, int, int, int);
-unsigned char *make_html_font_name(int);
-int get_real_font_size(int size);
+static int g_nobreak;
 
-struct style *get_style_by_ta(struct text_attrib *);
-void split_line_object(struct g_part *, struct g_object_text *, unsigned char *);
-void g_line_break(struct g_part *);
-void g_html_form_control(struct g_part *, struct form_control *);
-void do_image(struct g_part *, struct image_description *);
-void g_hr(struct g_part *p, struct hr_param *hr);
-void *g_html_special(struct g_part *, int, ...);
-void g_do_format(char *, char *, struct g_part *, unsigned char *);
-void g_scan_width(struct g_object **, int, int *);
-void g_scan_lines(struct g_object_line **, int, int *);
-int g_put_chars_conv(struct g_part *, unsigned char *, int);
-
-int g_nobreak;
-
-int get_real_font_size(int size)
+static int get_real_font_size(int size)
 {
 	int fs=d_opt->font_size;
 	
@@ -56,9 +37,7 @@ struct background *get_background(unsigned char *bg, unsigned char *bgcolor)
 	struct background *b;
 	struct rgb r;
 	b = mem_alloc(sizeof(struct background));
-	/* !!! FIXME: background image */
 	{
-		b->img = 0;
 		if (bgcolor && !decode_color(bgcolor, &r)) {
 			b->u.sRGB=(r.r << 16) + (r.g << 8) + r.b;
 		} else {
@@ -68,113 +47,41 @@ struct background *get_background(unsigned char *bg, unsigned char *bgcolor)
 	return b;
 }
 
-void g_put_chars(struct g_part *, unsigned char *, int);
-
-#define CH_BUF	256
-
-int g_put_chars_conv(struct g_part *p, unsigned char *c, int l)
-{
-	static char buffer[CH_BUF];
-	int bp = 0;
-	int pp = 0;
-	int total = 0;
-	if (format.attr & AT_GRAPHICS) {
-		g_put_chars(p, c, l);
-		return l;
-	}
-	/*{
-		debug("\"%.*s\"", l, c);
-	}*/
-	if (!l) g_put_chars(p, NULL, 0);
-	while (pp < l) {
-		unsigned char *e;
-		if (c[pp] < 128 && c[pp] != '&') {
-			put_c:
-			if (!(buffer[bp++] = c[pp++])) buffer[bp - 1] = ' ';
-			if (bp < CH_BUF) continue;
-			goto flush;
-		}
-		if (c[pp] != '&') {
-			struct conv_table *t;
-			int i;
-			if (!convert_table) goto put_c;
-			t = convert_table;
-			i = pp;
-			decode:
-			if (!t[c[i]].t) {
-				e = t[c[i]].u.str;
-			} else {
-				t = t[c[i++]].u.tbl;
-				if (i >= l) goto put_c;
-				goto decode;
-			}
-			pp = i + 1;
-		} else {
-			int i = pp + 1;
-			if (d_opt->plain & 1) goto put_c;
-			while (i < l && c[i] != ';' && c[i] != '&' && c[i] > ' ') i++;
-			if (!(e = get_entity_string(&c[pp + 1], i - pp - 1, d_opt->cp))) goto put_c;
-			pp = i + (i < l && c[i] == ';');
-		}
-		if (!e[0]) continue;
-		if (!e[1]) {
-			buffer[bp++] = e[0];
-			if (bp < CH_BUF) continue;
-			flush:
-			e = "";
-			goto flush1;
-		}
-		while (*e) {
-			buffer[bp++] = *(e++);
-			if (bp < CH_BUF) continue;
-			flush1:
-			g_put_chars(p, buffer, bp);
-			while (bp) if ((buffer[--bp] & 0xc0) != 0x80) total++;
-		}
-	}
-	if (bp) g_put_chars(p, buffer, bp);
-	while (bp) if ((buffer[--bp] & 0xc0) != 0x80) total++;
-	return total;
-}
+static void g_put_chars(void *, unsigned char *, int);
 
 /* Returns 0 to 2550 */
-int gray (int r, int g, int b)
+static int gray (int r, int g, int b)
 {
 	return r*3+g*6+b;
 }
 
 /* Tells if two colors are too near to be legible one on another */
 /* 1=too near 0=not too near */
-int too_near(int r1, int g1, int b1, int r2, int g2, int b2)
+static int too_near(int r1, int g1, int b1, int r2, int g2, int b2)
 {
-	int gray1,gray2;
+	int diff = abs(r1-r2) * 3 + abs(g1-g2) * 6 + abs(b1-b2);
 
-	gray1=gray(r1, g1, b1);
-	gray2=gray(r2, g2, b2);
-	if (gray1<=gray2)
-		return gray2-gray1<=400;
-	else
-		return gray1-gray2<=400;
+	return diff < 0x17f;
 }
 
 /* Fixes foreground based on background */
-void separate_fg_bg(int *fgr, int *fgg, int *fgb
+static void separate_fg_bg(int *fgr, int *fgg, int *fgb
 	, int bgr, int bgg, int bgb)
 {
-	if (too_near(*fgr, *fgg, *fgb, bgr, bgg, bgb)){
-		*fgr=255-bgr;
-		*fgg=255-bgg;
-		*fgb=255-bgb;
-	}else return;
-	if (too_near(*fgr, *fgg, *fgb, bgr, bgg, bgb)){
-		if (gray(bgr, bgg, bgb)<=1275)
-			*fgr=*fgg=*fgb=255;
+	if (too_near(*fgr, *fgg, *fgb, bgr, bgg, bgb)) {
+		*fgr = 255 - bgr;
+		*fgg = 255 - bgg;
+		*fgb = 255 - bgb;
+	} else return;
+	if (too_near(*fgr, *fgg, *fgb, bgr, bgg, bgb)) {
+		if (gray(bgr, bgg, bgb) <= 1275)
+			*fgr = *fgg = *fgb = 255;
 		else
-			*fgr=*fgg=*fgb=0;
+			*fgr = *fgg = *fgb = 0;
 	}
 }
 
-unsigned char *make_html_font_name(int attr)
+static unsigned char *make_html_font_name(int attr)
 {
 	unsigned char *str;
 	int len;
@@ -188,7 +95,7 @@ unsigned char *make_html_font_name(int attr)
 	return str;
 }
 
-struct style *get_style_by_ta(struct text_attrib *ta)
+static struct style *get_style_by_ta(struct text_attrib *ta)
 {
 	int fg_r,fg_g,fg_b; /* sRGB 0-255 values */
 	int fs = get_real_font_size(ta->fontsize);
@@ -328,22 +235,26 @@ void add_object_to_line(struct g_part *pp, struct g_object_line **lp, struct g_o
 	l->entries[l->n_entries - 1] = go;
 	*lp = l;
 	if (pp->cx == -1) pp->cx = par_format.leftmargin * G_HTML_MARGIN;
-	pp->cx += go->xw;
+	if (go->xw) {
+		pp->cx += pp->cx_w;
+		pp->cx_w = 0;
+		pp->cx += go->xw;
+	}
 }
 
 void flush_pending_text_to_line(struct g_part *p)
 {
-	if (!p->text) return;
-	add_object_to_line(p, &p->line, (struct g_object *)p->text);
+	struct g_object_text *t = p->text;
+	if (!t) return;
+	add_object_to_line(p, &p->line, (struct g_object *)t);
 	p->text = NULL;
 }
 
-void split_line_object(struct g_part *p, struct g_object_text *text, unsigned char *ptr)
+static void split_line_object(struct g_part *p, struct g_object_text *text, unsigned char *ptr)
 {
 	struct g_object_text *t2;
 	struct g_object_line *l2;
 	int n;
-	int esp;
 	if (!ptr) {
 		if (p->line && p->line->n_entries && (struct g_object *)text == p->line->entries[p->line->n_entries - 1]) {
 			flush_pending_line_to_obj(p, 0);
@@ -352,12 +263,15 @@ void split_line_object(struct g_part *p, struct g_object_text *text, unsigned ch
 		t2 = NULL;
 		goto nt2;
 	}
-	t2 = mem_calloc(sizeof(struct g_object_text) + strlen(ptr) + 1);
+#ifdef DEBUG
+	if (ptr < text->text || ptr >= text->text + strlen(text->text))
+		internal("split_line_object: split point (%p) pointing out of object (%p,%lx)", ptr, text->text, (unsigned long)strlen(text->text));
+#endif
+	t2 = mem_calloc(sizeof(struct g_object_text) + strlen(ptr));
 	t2->mouse_event = g_text_mouse;
 	t2->draw = g_text_draw;
 	t2->destruct = g_text_destruct;
 	t2->get_list = NULL;
-	esp = 1;
 	if (*ptr == ' ') {
 		strcpy(t2->text, ptr + 1);
 		*ptr = 0;
@@ -366,11 +280,9 @@ void split_line_object(struct g_part *p, struct g_object_text *text, unsigned ch
 		strcpy(t2->text, ptr);
 		ptr[0] = '-';
 		ptr[1] = 0;
-		esp = 0;
 	}
 	t2->y = text->y;
 	t2->style = g_clone_style(text->style);
-	t2->bg = text->bg;
 	t2->link_num = text->link_num;
 	t2->link_order = text->link_order;
 	text->xw = g_text_width(text->style, text->text);
@@ -419,14 +331,20 @@ void split_line_object(struct g_part *p, struct g_object_text *text, unsigned ch
 		p->text = t2;
 		p->pending_text_len = -1;
 		p->w.pos = t2->xw;
-		p->cx += g_char_width(t2->style, ' ');
+		p->cx_w = g_char_width(t2->style, ' ');
 	}
 	p->w.last_wrap = NULL;
 	p->w.last_wrap_obj = NULL;
 	t2 = p->text;
-	if (t2 && *t2->text && t2->text[strlen(t2->text) - 1] == ' ') {
-		p->w.last_wrap = &t2->text[strlen(t2->text) - 1];
-		p->w.last_wrap_obj = t2;
+	if (t2) {
+		int sl = strlen(t2->text);
+		if (sl >= 1 && t2->text[sl - 1] == ' ') {
+			p->w.last_wrap = &t2->text[sl - 1];
+			p->w.last_wrap_obj = t2;
+		} else if (sl >= 2 && t2->text[sl - 2] == 0xc2 && t2->text[sl - 1] == 0xad) {
+			p->w.last_wrap = &t2->text[sl - 2];
+			p->w.last_wrap_obj = t2;
+		}
 	}
 }
 
@@ -440,32 +358,38 @@ void add_object(struct g_part *p, struct g_object *o)
 	p->w.last_wrap = NULL;
 	p->w.last_wrap_obj = o;
 	p->w.pos += o->xw;
-	/*
-	if (p->w.pos > p->w.width && p->w.last_wrap_obj) {
-		split_line_object(p, p->w.last_wrap_obj, p->w.last_wrap);
-	}
-	*/
 }
 
-void g_line_break(struct g_part *p)
+static void g_line_break(void *p_)
 {
+	struct g_part *p = p_;
 	if (g_nobreak) {
 		g_nobreak = 0;
 		return;
 	}
 	flush_pending_text_to_line(p);
-	if (!p->line) {
+	if (!p->line || par_format.align == AL_NO) {
 		add_object_to_line(p, &p->line, NULL);
+		empty_line:
 		flush_pending_line_to_obj(p, get_real_font_size(format.fontsize));
-	} else /*if (p->w.pos || par_format.align == AL_NO)*/ {
+	} else {
+		int i;
+		for (i = 0; i < p->line->n_entries; i++) {
+			struct g_object *go = p->line->entries[i];
+			if (go->destruct != g_tag_destruct)
+				goto flush;
+		}
+		goto empty_line;
+		flush:
 		flush_pending_line_to_obj(p, 0);
 	}
 	if (p->cx > p->xmax) p->xmax = p->cx;
 	p->cx = -1;
+	p->cx_w = 0;
 }
 
 /* SHADOWED IN html_form_control */
-void g_html_form_control(struct g_part *p, struct form_control *fc)
+static void g_html_form_control(struct g_part *p, struct form_control *fc)
 {
 	if (!p->data) {
 		/*destroy_fc(fc);
@@ -497,7 +421,7 @@ void g_html_form_control(struct g_part *p, struct form_control *fc)
 	add_to_list(p->data->forms, fc);
 }
 
-struct link **putchars_link_ptr = NULL;
+static struct link **putchars_link_ptr = NULL;
 
 /* Probably releases clickable map */
 void release_image_map(struct image_map *map)
@@ -553,7 +477,7 @@ int is_in_area(struct map_area *a, int x, int y)
  * not specified. Autoscale is requested in im->autoscale.
  * If autoscale is specified, im->xsize and im->ysize must
  * be >0. */
-void do_image(struct g_part *p, struct image_description *im)
+static void do_image(struct g_part *p, struct image_description *im)
 {
 	struct g_object_image *io;
 	struct link *link;
@@ -656,7 +580,7 @@ void do_image(struct g_part *p, struct image_description *im)
 	ab:;
 }
 
-void g_hr(struct g_part *gp, struct hr_param *hr)
+static void g_hr(struct g_part *gp, struct hr_param *hr)
 {
 	unsigned char bgstr[8];
 	struct g_object_line *o;
@@ -677,11 +601,13 @@ void g_hr(struct g_part *gp, struct hr_param *hr)
 	add_object_to_line(gp, &gp->line, (struct g_object *)o);
 	line_breax = 0;
 	gp->cx = -1;
+	gp->cx_w = 0;
 }
 
 
-void *g_html_special(struct g_part *p, int c, ...)
+static void *g_html_special(void *p_, int c, ...)
 {
+	struct g_part *p = p_;
 	va_list l;
 	unsigned char *t;
 	struct form_control *fc;
@@ -697,7 +623,7 @@ void *g_html_special(struct g_part *p, int c, ...)
 			t = va_arg(l, unsigned char *);
 			va_end(l);
 			/* not needed to convert %AB here because html_tag will be called anyway */
-			tag = mem_calloc(sizeof(struct g_object_tag) + strlen(t) + 1);
+			tag = mem_calloc(sizeof(struct g_object_tag) + strlen(t));
 			tag->mouse_event = g_dummy_mouse;
 			tag->draw = g_dummy_draw;
 			tag->destruct = g_tag_destruct;
@@ -760,15 +686,34 @@ void *g_html_special(struct g_part *p, int c, ...)
 	return NULL;
 }
 
-unsigned char to_je_ale_prasarna[] = "";
+static unsigned char to_je_ale_prasarna[] = "";
 
-unsigned char *cached_font_face = to_je_ale_prasarna;
-struct text_attrib_beginning ta_cache = { -1, { 0, 0, 0, 0 }, { 0, 0, 0, 0 }, 0, 0 };
+static unsigned char *cached_font_face = to_je_ale_prasarna;
+static struct text_attrib_beginning ta_cache = { -1, { 0, 0, 0, 0 }, { 0, 0, 0, 0 }, 0, 0 };
 
-void g_put_chars(struct g_part *p, unsigned char *s, int l)
+static void g_put_chars(void *p_, unsigned char *s, int l)
 {
-	struct g_object_text *t = NULL;
+	struct g_part *p = p_;
+	struct g_object_text *t;
 	struct link *link;
+	int ptl;
+	unsigned char *sh;
+
+	if (l < 0) overalloc();
+
+again:
+	if (l > 2 && (sh = memchr(s + 1, 0xad, l - 2)) && sh[-1] == 0xc2) {
+		sh++;
+		g_put_chars(p_, s, sh - s);
+		l -= (sh - s);
+		s = sh;
+		goto again;
+	}
+
+	/*fprintf(stderr, "%d: '%.*s'\n", l, l, s);*/
+
+	t = NULL;
+
 	if (putchars_link_ptr) {
 		link = NULL;
 		goto check_link;
@@ -802,9 +747,7 @@ void g_put_chars(struct g_part *p, unsigned char *s, int l)
 		t->draw = g_text_draw;
 		t->destruct = g_text_destruct;
 		t->get_list = NULL;
-		/*t->style = get_style_by_ta(format);*/
 		t->style = g_clone_style(p->current_style);
-		t->bg = NULL; /* FIXME!!! */
 		t->yw = t->style->height;
 		/*t->xw = 0;
 		t->y = 0;*/
@@ -834,13 +777,17 @@ void g_put_chars(struct g_part *p, unsigned char *s, int l)
 	}
 	if (p->pending_text_len == -1) {
 		p->pending_text_len = strlen(p->text->text);
+		ptl = p->pending_text_len;
+		if (!ptl) ptl = 1;
 		goto a1;
 	}
-	if ((p->pending_text_len & ~(ALLOC_GR - 1)) != ((p->pending_text_len + l) & ~(ALLOC_GR - 1))) a1:{
+	ptl = p->pending_text_len;
+	if (!ptl) ptl = 1;
+	if (((ptl + ALLOC_GR - 1) & ~(ALLOC_GR - 1)) != ((ptl + l + ALLOC_GR - 1) & ~(ALLOC_GR - 1))) a1: {
 		struct g_object_text *t;
 		if ((unsigned)l > MAXINT) overalloc();
-		if ((unsigned)p->pending_text_len + (unsigned)l > MAXINT - ALLOC_GR) overalloc();
-		t = mem_realloc(p->text, sizeof(struct g_object_text) + ((p->pending_text_len + l + ALLOC_GR) & ~(ALLOC_GR - 1)));
+		if ((unsigned)ptl + (unsigned)l > MAXINT - ALLOC_GR) overalloc();
+		t = mem_realloc(p->text, sizeof(struct g_object_text) + ((ptl + l + ALLOC_GR - 1) & ~(ALLOC_GR - 1)));
 		if (p->w.last_wrap >= p->text->text && p->w.last_wrap < p->text->text + p->pending_text_len) p->w.last_wrap += (unsigned char *)t - (unsigned char *)p->text;
 		if (p->w.last_wrap_obj == p->text) p->w.last_wrap_obj = t;
 		p->text = t;
@@ -852,6 +799,7 @@ void g_put_chars(struct g_part *p, unsigned char *s, int l)
 		p->w.style = p->text->style;
 		p->w.obj = p->text;
 		p->w.width = rm(par_format) - par_format.leftmargin * G_HTML_MARGIN;
+		p->w.force_break = 0;
 		if (p->w.width < 0) p->w.width = 0;
 		if (!g_wrap_text(&p->w)) {
 			split_line_object(p, p->w.last_wrap_obj, p->w.last_wrap);
@@ -928,10 +876,10 @@ void g_put_chars(struct g_part *p, unsigned char *s, int l)
 	goto back_link;
 }
 
-void g_do_format(char *start, char *end, struct g_part *part, unsigned char *head)
+static void g_do_format(unsigned char *start, unsigned char *end, struct g_part *part, unsigned char *head)
 {
 	pr(
-	parse_html(start, end, (int (*)(void *, unsigned char *, int)) g_put_chars_conv, (void (*)(void *)) g_line_break, (void *(*)(void *, int, ...)) g_html_special, part, head);
+	parse_html(start, end, (void (*)(void *, unsigned char *, int))g_put_chars, g_line_break, (void *(*)(void *, int, ...))g_html_special, part, head);
 	/*if ((part->y -= line_breax) < 0) part->y = 0;*/
 	flush_pending_text_to_line(part);
 	flush_pending_line_to_obj(part, 0);
@@ -950,7 +898,7 @@ struct g_table_cache_entry {
 	struct g_part p;
 };
 
-struct list_head g_table_cache = { &g_table_cache, &g_table_cache };
+static struct list_head g_table_cache = { &g_table_cache, &g_table_cache };
 
 void g_free_table_cache(void)
 {
@@ -998,9 +946,7 @@ struct g_part *g_format_html_part(unsigned char *start, unsigned char *end, int
 		a->draw = g_area_draw;
 		a->destruct = g_area_destruct;
 		a->get_list = g_area_get_list;
-		a->lfo = DUMMY;
-		a->rfo = DUMMY;
-		a->n_lines = 0;
+		/*a->n_lines = 0;*/
 		p->root = a;
 		init_list(p->uf);
 	}
@@ -1019,7 +965,8 @@ struct g_part *g_format_html_part(unsigned char *start, unsigned char *end, int
 	par_format.list_number = 0;
 	par_format.dd_margin = 0;
 	p->cx = -1;
-	g_nobreak = 1;
+	p->cx_w = 0;
+	g_nobreak = align != AL_NO;
 	g_do_format(start, end, p, head);
 	g_nobreak = 0;
 	line_breax = 1;
@@ -1034,7 +981,7 @@ struct g_part *g_format_html_part(unsigned char *start, unsigned char *end, int
 
 	wa = g_get_area_width(p->root);
 	if (wa > p->x) p->x = wa;
-	g_x_extend_area(p->root, p->x, 0);
+	g_x_extend_area(p->root, p->x, 0, align);
 	if (p->x > p->xmax) p->xmax = p->x;
 	p->y = p->root->yw;
 	/*debug("WIDTH: obj (%d, %d), p (%d %d)", p->root->xw, p->root->yw, p->x, p->y);*/
@@ -1080,15 +1027,7 @@ void g_release_part(struct g_part *p)
 	if (p->current_style) g_free_style(p->current_style);
 }
 
-void g_scan_width(struct g_object **o, int n, int *w)
-{
-	while (n--) {
-		if ((*o)->x + (*o)->xw > *w) *w = (*o)->x + (*o)->xw;
-		o++;
-	}
-}
-
-void g_scan_lines(struct g_object_line **o, int n, int *w)
+static void g_scan_lines(struct g_object_line **o, int n, int *w)
 {
 	while (n--) {
 		if ((*o)->n_entries) {
@@ -1102,13 +1041,11 @@ void g_scan_lines(struct g_object_line **o, int n, int *w)
 int g_get_area_width(struct g_object_area *a)
 {
 	int w = 0;
-	g_scan_width(a->lfo, a->n_lfo, &w);
-	g_scan_width(a->rfo, a->n_rfo, &w);
 	g_scan_lines(a->lines, a->n_lines, &w);
 	return w;
 }
 
-void g_x_extend_area(struct g_object_area *a, int width, int height)
+void g_x_extend_area(struct g_object_area *a, int width, int height, int align)
 {
 	struct g_object_line *l;
 	int i;
@@ -1116,7 +1053,7 @@ void g_x_extend_area(struct g_object_area *a, int width, int height)
 	for (i = 0; i < a->n_lines; i++) {
 		a->lines[i]->xw = width;
 	}
-	for (i = a->n_lines - 1; i >= 0; i--) {
+	if (align != AL_NO) for (i = a->n_lines - 1; i >= 0; i--) {
 		l = a->lines[i];
 		if (!l->n_entries) {
 			a->yw -= l->yw;
