@@ -43,7 +43,7 @@
 
 #if defined(UNIX)
 
-static inline int dir_sep(char x) { return x == '/'; }
+static inline int dir_sep(unsigned char x) { return x == '/'; }
 #define NEWLINE "\n"
 #define FS_UNIX_RIGHTS
 #define FS_UNIX_HARDLINKS
@@ -64,7 +64,7 @@ static inline int dir_sep(char x) { return x == '/'; }
 
 #elif defined(OS2)
 
-static inline int dir_sep(char x) { return x == '/' || x == '\\'; }
+static inline int dir_sep(unsigned char x) { return x == '/' || x == '\\'; }
 #define NEWLINE "\r\n"
 /*#define NO_ASYNC_LOOKUP*/
 #define SYSTEM_ID SYS_OS2
@@ -81,7 +81,7 @@ static inline int dir_sep(char x) { return x == '/' || x == '\\'; }
 
 #elif defined(WIN32)
 
-static inline int dir_sep(char x) { return x == '/' || x == '\\'; }
+static inline int dir_sep(unsigned char x) { return x == '/' || x == '\\'; }
 #define NEWLINE "\r\n"
 /*#define NO_ASYNC_LOOKUP*/
 #define SYSTEM_ID SYS_WIN_32
@@ -93,17 +93,24 @@ static inline int dir_sep(char x) { return x == '/' || x == '\\'; }
 #define DOS_FS
 #define SET_WINDOW_TITLE_UTF_8
 #define ASSOC_CONS_XWIN
+#ifdef _UWIN
+#define DISABLE_SMB
+#endif
+#ifdef __CYGWIN__
 #define OS_BAD_SIGNALS
+#endif
 #ifndef HAVE_PTHREADS
 #define HAVE_PTHREADS
 #endif
-#ifdef HAVE_SYS_UN_H
+#if defined(HAVE_SYS_UN_H) && !defined(_UWIN)
 #define USE_AF_UNIX
+#else
+#define DONT_USE_AF_UNIX
 #endif
 
 #elif defined(INTERIX)
 
-static inline int dir_sep(char x) { return x == '/'; }
+static inline int dir_sep(unsigned char x) { return x == '/'; }
 #define NEWLINE "\n"
 #define FS_UNIX_RIGHTS
 #define FS_UNIX_HARDLINKS
@@ -124,7 +131,7 @@ static inline int dir_sep(char x) { return x == '/'; }
 
 #elif defined(BEOS)
 
-static inline int dir_sep(char x) { return x == '/'; }
+static inline int dir_sep(unsigned char x) { return x == '/'; }
 #define NEWLINE "\n"
 #define NO_ASYNC_LOOKUP /* async lookup works on BeOS but crashes the Haiku kernel */
 #define FS_UNIX_RIGHTS
@@ -150,7 +157,7 @@ static inline int dir_sep(char x) { return x == '/'; }
 
 #elif defined(RISCOS)
 
-static inline int dir_sep(char x) { return x == '/' || x == '\\'; }
+static inline int dir_sep(unsigned char x) { return x == '/' || x == '\\'; }
 #define NEWLINE "\n"
 #define SYSTEM_ID SYS_RISCOS
 #define SYSTEM_NAME "RISC OS"
@@ -163,7 +170,7 @@ static inline int dir_sep(char x) { return x == '/' || x == '\\'; }
 
 #elif defined(ATHEOS)
 
-static inline int dir_sep(char x) { return x == '/'; }
+static inline int dir_sep(unsigned char x) { return x == '/'; }
 #define NEWLINE "\n"
 #define FS_UNIX_RIGHTS
 #define FS_UNIX_HARDLINKS
@@ -178,7 +185,7 @@ static inline int dir_sep(char x) { return x == '/'; }
 
 #elif defined(SPAD)
 
-static inline int dir_sep(char x) { return x == '/'; }
+static inline int dir_sep(unsigned char x) { return x == '/'; }
 #define NEWLINE "\n"
 #define SYSTEM_ID SYS_SPAD
 #define SYSTEM_NAME "Spad"
