@@ -24,16 +24,10 @@ static struct list_head h_conns = {&h_conns, &h_conns};
 struct list_head keepalive_connections = {&keepalive_connections, &keepalive_connections};
 
 /* prototypes */
-struct k_conn *is_host_on_keepalive_list(struct connection *);
+static void send_connection_info(struct connection *c);
 static void check_keepalive_connections(void);
 
-
-void no_owner(void)
-{
-	internal("connection has no owner");
-}
-
-long connect_info(int type)
+unsigned long connect_info(int type)
 {
 	int i = 0;
 	struct connection *ce;
@@ -49,12 +43,13 @@ long connect_info(int type)
 			foreach(ce, queue) i += ce->state == S_TRANS;
 			return i;
 		case CI_KEEP:
+			check_keepalive_connections();
 			foreach(cee, keepalive_connections) i++;
 			return i;
 		case CI_LIST:
 			return (long) &queue;
 		default:
-			internal("cache_info: bad request");
+			internal("connect_info: bad request");
 	}
 	return 0;
 }
@@ -84,25 +79,25 @@ static int st_r = 0;
 static void stat_timer(struct connection *c)
 {
 	struct remaining_info *r = &c->prg;
-	ttime a = get_time() - r->last_time;
+	ttime a = (uttime)get_time() - (uttime)r->last_time;
 	if (getpri(c) == PRI_CANCEL && (c->est_length > (longlong)memory_cache_size * MAX_CACHED_OBJECT || c->from > (longlong)memory_cache_size * MAX_CACHED_OBJECT)) register_bottom_half(check_queue, NULL);
 	if (c->state > S_WAIT) {
 		r->loaded = c->received;
 		if ((r->size = c->est_length) < (r->pos = c->from) && r->size != -1)
 			r->size = c->from;
-		r->dis_b += a;
+		r->dis_b += (uttime)a;
 		while (r->dis_b >= SPD_DISP_TIME * CURRENT_SPD_SEC) {
 			r->cur_loaded -= r->data_in_secs[0];
 			memmove(r->data_in_secs, r->data_in_secs + 1, sizeof(int) * (CURRENT_SPD_SEC - 1));
 			r->data_in_secs[CURRENT_SPD_SEC - 1] = 0;
-			r->dis_b -= SPD_DISP_TIME;
+			r->dis_b -= (uttime)SPD_DISP_TIME;
 		}
 		r->data_in_secs[CURRENT_SPD_SEC - 1] += r->loaded - r->last_loaded;
-		r->cur_loaded += r->loaded - r->last_loaded;
-		r->last_loaded = r->loaded;
-		r->elapsed += a;
+		r->cur_loaded += (uttime)r->loaded - (uttime)r->last_loaded;
+		r->last_loaded = (uttime)r->loaded;
+		r->elapsed += (uttime)a;
 	}
-	r->last_time += a;
+	r->last_time += (uttime)a;
 	r->timer = install_timer(SPD_DISP_TIME, (void (*)(void *))stat_timer, c);
 	if (!st_r) send_connection_info(c);
 }
@@ -137,7 +132,7 @@ void setcstate(struct connection *c, int state)
 	if (state >= 0) send_connection_info(c);
 }
 
-struct k_conn *is_host_on_keepalive_list(struct connection *c)
+static struct k_conn *is_host_on_keepalive_list(struct connection *c)
 {
 	unsigned char *ho;
 	int po;
@@ -159,19 +154,25 @@ int get_keepalive_socket(struct connection *c)
 {
 	struct k_conn *k;
 	int cc;
+	if (c->tries > 0 || c->unrestartable) return -1;
 	if (!(k = is_host_on_keepalive_list(c))) return -1;
 	cc = k->conn;
 	del_from_list(k);
 	mem_free(k->host);
 	mem_free(k);
 	c->sock1 = cc;
+	if (max_tries == 1) c->tries = -1;
 	return 0;
 }
 
 void abort_all_keepalive_connections(void)
 {
 	struct k_conn *k;
-	foreach(k, keepalive_connections) mem_free(k->host), close(k->conn);
+	int rs;
+	foreach(k, keepalive_connections) {
+		mem_free(k->host);
+		EINTRLOOP(rs, close(k->conn));
+	}
 	free_list(keepalive_connections);
 	check_keepalive_connections();
 }
@@ -179,13 +180,14 @@ void abort_all_keepalive_connections(void)
 static void free_connection_data(struct connection *c)
 {
 	struct h_conn *h;
+	int rs;
 	if (c->sock1 != -1) set_handlers(c->sock1, NULL, NULL, NULL, NULL);
 	if (c->sock2 != -1) set_handlers(c->sock2, NULL, NULL, NULL, NULL);
 	close_socket(&c->sock2);
 	if (c->pid) {
-		kill(c->pid, SIGINT);
-		kill(c->pid, SIGTERM);
-		kill(c->pid, SIGKILL);
+		EINTRLOOP(rs, kill(c->pid, SIGINT));
+		EINTRLOOP(rs, kill(c->pid, SIGTERM));
+		EINTRLOOP(rs, kill(c->pid, SIGKILL));
 		c->pid = 0;
 	}
 	if (!c->running) {
@@ -221,7 +223,7 @@ static void free_connection_data(struct connection *c)
 	}
 }
 
-void send_connection_info(struct connection *c)
+static void send_connection_info(struct connection *c)
 {
 	int st = c->state;
 	tcount count = c->count;
@@ -236,20 +238,31 @@ void send_connection_info(struct connection *c)
 
 static void del_connection(struct connection *c)
 {
+	struct cache_entry *ce = c->cache;
+	if (ce) ce->refcount++;
 	del_from_list(c);
 	send_connection_info(c);
+	if (ce) ce->refcount--;
+	if (c->detached) {
+		if (ce && !ce->url[0] && !is_entry_used(ce) && !ce->refcount)
+			delete_cache_entry(ce);
+	} else {
+		if (ce)
+			trim_cache_entry(ce);
+	}
 	mem_free(c->url);
 	if (c->prev_url) mem_free(c->prev_url);
 	mem_free(c);
 }
 
 #ifdef DEBUG
-void check_queue_bugs(void);
+static void check_queue_bugs(void);
 #endif
 
 void add_keepalive_socket(struct connection *c, ttime timeout)
 {
 	struct k_conn *k;
+	int rs;
 	free_connection_data(c);
 	if (c->sock1 == -1) {
 		internal("keepalive connection not connected");
@@ -273,7 +286,7 @@ void add_keepalive_socket(struct connection *c, ttime timeout)
 	register_bottom_half(check_queue, NULL);
 	return;
 	close:
-	close(c->sock1);
+	EINTRLOOP(rs, close(c->sock1));
 #ifdef DEBUG
 	check_queue_bugs();
 #endif
@@ -282,8 +295,9 @@ void add_keepalive_socket(struct connection *c, ttime timeout)
 
 static void del_keepalive_socket(struct k_conn *kc)
 {
+	int rs;
 	del_from_list(kc);
-	close(kc->conn);
+	EINTRLOOP(rs, close(kc->conn));
 	mem_free(kc->host);
 	mem_free(kc);
 }
@@ -296,13 +310,13 @@ static void keepalive_timer(void *x)
 	check_keepalive_connections();
 }
 
-void check_keepalive_connections(void)
+static void check_keepalive_connections(void)
 {
 	struct k_conn *kc;
 	ttime ct = get_time();
 	int p = 0;
 	if (keepalive_timeout != -1) kill_timer(keepalive_timeout), keepalive_timeout = -1;
-	foreach(kc, keepalive_connections) if (can_read(kc->conn) || ct - kc->add_time > kc->timeout) {
+	foreach(kc, keepalive_connections) if (can_read(kc->conn) || (uttime)ct - (uttime)kc->add_time > (uttime)kc->timeout) {
 		kc = kc->prev;
 		del_keepalive_socket(kc->next);
 	} else p++;
@@ -341,12 +355,13 @@ static void sort_queue(void)
 static void interrupt_connection(struct connection *c)
 {
 #ifdef HAVE_SSL
-	if (c->ssl == (void *)-1) c->ssl = 0;
-	if(c->ssl) {
+	if (c->ssl == (void *)-1) c->ssl = NULL;
+	if (c->ssl) {
 		SSL_free(c->ssl);
-		c->ssl=NULL;
+		c->ssl = NULL;
 	}
 #endif
+	if (c->sock1 != -1) set_handlers(c->sock1, NULL, NULL, NULL, NULL);
 	close_socket(&c->sock1);
 	free_connection_data(c);
 }
@@ -380,7 +395,7 @@ static int try_to_suspend_connection(struct connection *c, unsigned char *ho)
 	return -1;
 }
 
-void run_connection(struct connection *c)
+static void run_connection(struct connection *c)
 {
 	struct h_conn *hc;
 	void (*func)(struct connection *);
@@ -425,6 +440,30 @@ void run_connection(struct connection *c)
 	func(c);
 }
 
+static int is_connection_seekable(struct connection *c)
+{
+	unsigned char *protocol = get_protocol_name(c->url);
+	if (!strcasecmp(protocol, "http") || !strcasecmp(protocol, "https") ||
+	    !strcasecmp(protocol, "proxy")) {
+		unsigned char *d;
+		mem_free(protocol);
+		if (!c->cache || !c->cache->head)
+			return 1;
+		d = parse_http_header(c->cache->head, "Accept-Ranges", NULL);
+		if (d) {
+			mem_free(d);
+			return 1;
+		}
+		return 0;
+	}
+	if (!strcasecmp(protocol, "ftp")) {
+		mem_free(protocol);
+		return 1;
+	}
+	mem_free(protocol);
+	return 0;
+}
+
 int is_connection_restartable(struct connection *c)
 {
 	return !(c->unrestartable >= 2 || (c->tries + 1 >= (max_tries ? max_tries : 1000)));
@@ -443,7 +482,6 @@ void retry_connection(struct connection *c)
 {
 	interrupt_connection(c);
 	if (!is_connection_restartable(c)) {
-		/*send_connection_info(c);*/
 		del_connection(c);
 #ifdef DEBUG
 		check_queue_bugs();
@@ -459,7 +497,6 @@ void retry_connection(struct connection *c)
 void abort_connection(struct connection *c)
 {
 	if (c->running) interrupt_connection(c);
-	/*send_connection_info(c);*/
 	del_connection(c);
 #ifdef DEBUG
 	check_queue_bugs();
@@ -485,7 +522,7 @@ static int try_connection(struct connection *c)
 }
 
 #ifdef DEBUG
-void check_queue_bugs(void)
+static void check_queue_bugs(void)
 {
 	struct connection *d;
 	int p = 0, ps = 0;
@@ -495,14 +532,16 @@ void check_queue_bugs(void)
 	foreach(d, queue) {
 		int q = getpri(d);
 		cc += d->running;
-		if (q < p) if (!ps) {
-			internal("queue is not sorted");
-			sort_queue();
-			ps = 1;
-			goto again;
-		} else {
-			internal("queue is not sorted even after sort_queue!");
-			break;
+		if (q < p) {
+			if (!ps) {
+				internal("queue is not sorted");
+				sort_queue();
+				ps = 1;
+				goto again;
+			} else {
+				internal("queue is not sorted even after sort_queue!");
+				break;
+			}
 		} else p = q;
 		if (d->state < 0) {
 			internal("interrupted connection on queue (conn %s, state %d)", d->url, d->state);
@@ -577,11 +616,12 @@ unsigned char *get_proxy(unsigned char *url)
 
 /* prev_url is a pointer to previous url or NULL */
 /* prev_url will NOT be deallocated */
-void load_url(unsigned char *url, unsigned char * prev_url, struct status *stat, int pri, int no_cache)
+void load_url(unsigned char *url, unsigned char *prev_url, struct status *stat, int pri, int no_cache, int no_compress, off_t position)
 {
 	struct cache_entry *e = NULL;
 	struct connection *c;
 	unsigned char *u;
+	int must_detach = 0;
 	if (stat) stat->c = NULL, stat->ce = NULL, stat->pri = pri;
 #ifdef DEBUG
 	foreach(c, queue) {
@@ -597,18 +637,34 @@ void load_url(unsigned char *url, unsigned char * prev_url, struct status *stat,
 	}
 #endif
 	if (stat) stat->state = S_OUT_OF_MEM, stat->prev_error = 0;
-	if (no_cache <= NC_CACHE && !find_in_cache(url, &e) && !e->incomplete) {
+	if (no_cache <= NC_CACHE && !find_in_cache(url, &e)) {
+		if (e->incomplete) {
+			e->refcount--;
+			goto skip_cache;
+		}
 		if (!aggressive_cache && no_cache > NC_ALWAYS_CACHE) {
 			if (e->expire_time && e->expire_time < time(NULL)) {
 				if (no_cache < NC_IF_MOD) no_cache = NC_IF_MOD;
+				e->refcount--;
+				goto skip_cache;
+			}
+		}
+		if (no_compress) {
+			unsigned char *enc;
+			enc = parse_http_header(e->head, "Content-Encoding", NULL);
+			if (enc) {
+				mem_free(enc);
+				e->refcount--;
+				must_detach = 1;
 				goto skip_cache;
 			}
 		}
 		if (stat) {
 			stat->ce = e;
-			stat->state = S_OK;
+			stat->state = S__OK;
 			if (stat->end) stat->end(stat, stat->data);
 		}
+		e->refcount--;
 		return;
 	}
 	skip_cache:
@@ -621,6 +677,20 @@ void load_url(unsigned char *url, unsigned char * prev_url, struct status *stat,
 	}
 	u = get_proxy(url);
 	foreach(c, queue) if (!c->detached && !strcmp(c->url, u)) {
+		if (c->from < position) continue;
+		if (no_compress && !c->no_compress) {
+			unsigned char *enc;
+			if ((c->state >= S_WAIT && c->state < S_TRANS) || !c->cache) {
+				must_detach = 1;
+				break;
+			}
+			enc = parse_http_header(c->cache->head, "Content-Encoding", NULL);
+			if (enc) {
+				mem_free(enc);
+				must_detach = 1;
+				break;
+			}
+		}
 		mem_free(u);
 		if (getpri(c) > pri) {
 			del_from_list(c);
@@ -646,7 +716,20 @@ void load_url(unsigned char *url, unsigned char * prev_url, struct status *stat,
 	c->prev_url = stracpy(prev_url);
 	c->running = 0;
 	c->prev_error = 0;
-	c->from = no_cache >= NC_IF_MOD || !e || e->frag.next == &e->frag || ((struct fragment *)e->frag.next)->offset ? 0 : ((struct fragment *)e->frag.next)->length;
+	if (position || must_detach) {
+		c->from = position;
+	} else if (no_cache >= NC_IF_MOD || !e) {
+		c->from = 0;
+	} else {
+		struct fragment *frag;
+		c->from = 0;
+		foreach(frag, e->frag) {
+			if (frag->offset != c->from)
+				break;
+			c->from += frag->length;
+		}
+		
+	}
 	memset(c->pri, 0, sizeof c->pri);
 	c->pri[pri] = 1;
 	c->no_cache = no_cache;
@@ -660,8 +743,28 @@ void load_url(unsigned char *url, unsigned char * prev_url, struct status *stat,
 	init_list(c->statuss);
 	c->est_length = -1;
 	c->unrestartable = 0;
+#ifdef HAVE_ANY_COMPRESSION
+	c->no_compress = http_options.no_compression || no_compress;
+#else
+	c->no_compress = 1;
+#endif
 	c->prg.timer = -1;
 	c->timer = -1;
+	if (position || must_detach) {
+		if (new_cache_entry(c->url, &c->cache)) {
+			mem_free(c->url);
+			if (c->prev_url) mem_free(c->prev_url);
+			mem_free(c);
+			if (stat) {
+				stat->state = S_OUT_OF_MEM;
+				if (stat->end) stat->end(stat, stat->data);
+			}
+			return;
+		}
+		c->cache->refcount--;
+		detach_cache_entry(c->cache);
+		c->detached = 2;
+	}
 	if (stat) {
 		stat->prg = &c->prg;
 		stat->c = c;
@@ -687,10 +790,13 @@ void change_connection(struct status *oldstat, struct status *newstat, int newpr
 	oldpri = oldstat->pri;
 	if (oldstat->state < 0) {
 		if (newstat) {
+			struct cache_entry *ce = oldstat->ce;
+			if (ce) ce->refcount++;
 			newstat->ce = oldstat->ce;
 			newstat->state = oldstat->state;
 			newstat->prev_error = oldstat->prev_error;
 			if (newstat->end) newstat->end(newstat, newstat->data);
+			if (ce) ce->refcount--;
 		}
 		return;
 	}
@@ -732,23 +838,31 @@ void detach_connection(struct status *stat, off_t pos)
 	off_t l;
 	if (stat->state < 0) return;
 	c = stat->c;
-	if (c->detached) goto detach_done;
 	if (!c->cache) return;
+	if (c->detached) goto detach_done;
 	if (c->est_length == -1) l = c->from;
 	else l = c->est_length;
-	if (l < (longlong)memory_cache_size * MAX_CACHED_OBJECT) return;
+	if (l < (longlong)memory_cache_size * MAX_CACHED_OBJECT && !(pos > c->from && is_connection_seekable(c))) return;
 	l = 0;
 	for (i = 0; i < PRI_CANCEL; i++) l += c->pri[i];
 	if (!l) internal("detaching free connection");
 	if (l != 1 || c->cache->refcount) return;
 	shrink_memory(SH_CHECK_QUOTA);
-	c->cache->url[0] = 0;
+	detach_cache_entry(c->cache);
 	c->detached = 1;
 	detach_done:
 	free_entry_to(c->cache, pos);
+
+	if (c->detached < 2 && pos > c->from && is_connection_seekable(c)) {
+		int running = c->running;
+		if (running) interrupt_connection(c);
+		c->from = pos;
+		if (running) run_connection(c);
+		c->detached = 2;
+	}
 }
 
-void connection_timeout(struct connection *c)
+static void connection_timeout(struct connection *c)
 {
 	c->timer = -1;
 	setcstate(c, S_TIMEOUT);
@@ -813,7 +927,7 @@ struct blacklist_entry {
 	unsigned char host[1];
 };
 
-struct list_head blacklist = { &blacklist, &blacklist };
+static struct list_head blacklist = { &blacklist, &blacklist };
 
 void add_blacklist_entry(unsigned char *host, int flags)
 {
@@ -854,58 +968,59 @@ void free_blacklist(void)
 }
 
 struct s_msg_dsc msg_dsc[] = {
-	{S_WAIT,		TEXT(T_WAITING_IN_QUEUE)},
-	{S_DNS,			TEXT(T_LOOKING_UP_HOST)},
-	{S_CONN,		TEXT(T_MAKING_CONNECTION)},
-	{S_SOCKS_NEG,		TEXT(T_SOCKS_NEGOTIATION)},
-	{S_SSL_NEG,		TEXT(T_SSL_NEGOTIATION)},
-	{S_SENT,		TEXT(T_REQUEST_SENT)},
-	{S_LOGIN,		TEXT(T_LOGGING_IN)},
-	{S_GETH,		TEXT(T_GETTING_HEADERS)},
-	{S_PROC,		TEXT(T_SERVER_IS_PROCESSING_REQUEST)},
-	{S_TRANS,		TEXT(T_TRANSFERRING)},
+	{S_WAIT,		TEXT_(T_WAITING_IN_QUEUE)},
+	{S_DNS,			TEXT_(T_LOOKING_UP_HOST)},
+	{S_CONN,		TEXT_(T_MAKING_CONNECTION)},
+	{S_SOCKS_NEG,		TEXT_(T_SOCKS_NEGOTIATION)},
+	{S_SSL_NEG,		TEXT_(T_SSL_NEGOTIATION)},
+	{S_SENT,		TEXT_(T_REQUEST_SENT)},
+	{S_LOGIN,		TEXT_(T_LOGGING_IN)},
+	{S_GETH,		TEXT_(T_GETTING_HEADERS)},
+	{S_PROC,		TEXT_(T_SERVER_IS_PROCESSING_REQUEST)},
+	{S_TRANS,		TEXT_(T_TRANSFERRING)},
 
-	{S_WAIT_REDIR,		TEXT(T_WAITING_FOR_REDIRECT_CONFIRMATION)},
-	{S_OK,			TEXT(T_OK)},
-	{S_INTERRUPTED,		TEXT(T_INTERRUPTED)},
-	{S_EXCEPT,		TEXT(T_SOCKET_EXCEPTION)},
-	{S_INTERNAL,		TEXT(T_INTERNAL_ERROR)},
-	{S_OUT_OF_MEM,		TEXT(T_OUT_OF_MEMORY)},
-	{S_NO_DNS,		TEXT(T_HOST_NOT_FOUND)},
-	{S_CANT_WRITE,		TEXT(T_ERROR_WRITING_TO_SOCKET)},
-	{S_CANT_READ,		TEXT(T_ERROR_READING_FROM_SOCKET)},
-	{S_MODIFIED,		TEXT(T_DATA_MODIFIED)},
-	{S_BAD_URL,		TEXT(T_BAD_URL_SYNTAX)},
-	{S_TIMEOUT,		TEXT(T_RECEIVE_TIMEOUT)},
-	{S_RESTART,		TEXT(T_REQUEST_MUST_BE_RESTARTED)},
-	{S_STATE,		TEXT(T_CANT_GET_SOCKET_STATE)},
-	{S_CYCLIC_REDIRECT,	TEXT(T_CYCLIC_REDIRECT)},
-	{S_LARGE_FILE,		TEXT(T_TOO_LARGE_FILE)},
+	{S__OK,			TEXT_(T_OK)},
+	{S_INTERRUPTED,		TEXT_(T_INTERRUPTED)},
+	{S_EXCEPT,		TEXT_(T_SOCKET_EXCEPTION)},
+	{S_INTERNAL,		TEXT_(T_INTERNAL_ERROR)},
+	{S_OUT_OF_MEM,		TEXT_(T_OUT_OF_MEMORY)},
+	{S_NO_DNS,		TEXT_(T_HOST_NOT_FOUND)},
+	{S_CANT_WRITE,		TEXT_(T_ERROR_WRITING_TO_SOCKET)},
+	{S_CANT_READ,		TEXT_(T_ERROR_READING_FROM_SOCKET)},
+	{S_MODIFIED,		TEXT_(T_DATA_MODIFIED)},
+	{S_BAD_URL,		TEXT_(T_BAD_URL_SYNTAX)},
+	{S_TIMEOUT,		TEXT_(T_RECEIVE_TIMEOUT)},
+	{S_RESTART,		TEXT_(T_REQUEST_MUST_BE_RESTARTED)},
+	{S_STATE,		TEXT_(T_CANT_GET_SOCKET_STATE)},
+	{S_CYCLIC_REDIRECT,	TEXT_(T_CYCLIC_REDIRECT)},
+	{S_LARGE_FILE,		TEXT_(T_TOO_LARGE_FILE)},
 
-	{S_HTTP_ERROR,		TEXT(T_BAD_HTTP_RESPONSE)},
-	{S_HTTP_100,		TEXT(T_HTTP_100)},
-	{S_HTTP_204,		TEXT(T_NO_CONTENT)},
+	{S_HTTP_ERROR,		TEXT_(T_BAD_HTTP_RESPONSE)},
+	{S_HTTP_100,		TEXT_(T_HTTP_100)},
+	{S_HTTP_204,		TEXT_(T_NO_CONTENT)},
 
-	{S_FILE_TYPE,		TEXT(T_UNKNOWN_FILE_TYPE)},
-	{S_FILE_ERROR,		TEXT(T_ERROR_OPENING_FILE)},
+	{S_FILE_TYPE,		TEXT_(T_UNKNOWN_FILE_TYPE)},
+	{S_FILE_ERROR,		TEXT_(T_ERROR_OPENING_FILE)},
 
-	{S_FTP_ERROR,		TEXT(T_BAD_FTP_RESPONSE)},
-	{S_FTP_UNAVAIL,		TEXT(T_FTP_SERVICE_UNAVAILABLE)},
-	{S_FTP_LOGIN,		TEXT(T_BAD_FTP_LOGIN)},
-	{S_FTP_PORT,		TEXT(T_FTP_PORT_COMMAND_FAILED)},
-	{S_FTP_NO_FILE,		TEXT(T_FILE_NOT_FOUND)},
-	{S_FTP_FILE_ERROR,	TEXT(T_FTP_FILE_ERROR)},
+	{S_FTP_ERROR,		TEXT_(T_BAD_FTP_RESPONSE)},
+	{S_FTP_UNAVAIL,		TEXT_(T_FTP_SERVICE_UNAVAILABLE)},
+	{S_FTP_LOGIN,		TEXT_(T_BAD_FTP_LOGIN)},
+	{S_FTP_PORT,		TEXT_(T_FTP_PORT_COMMAND_FAILED)},
+	{S_FTP_NO_FILE,		TEXT_(T_FILE_NOT_FOUND)},
+	{S_FTP_FILE_ERROR,	TEXT_(T_FTP_FILE_ERROR)},
 
-	{S_SSL_ERROR,		TEXT(T_SSL_ERROR)},
-	{S_NO_SSL,		TEXT(T_NO_SSL)},
-	{S_BAD_SOCKS_VERSION,	TEXT(T_BAD_SOCKS_VERSION)},
-	{S_SOCKS_REJECTED,	TEXT(T_SOCKS_REJECTED_OR_FAILED)},
-	{S_SOCKS_NO_IDENTD,	TEXT(T_SOCKS_NO_IDENTD)},
-	{S_SOCKS_BAD_USERID,	TEXT(T_SOCKS_BAD_USERID)},
-	{S_SOCKS_UNKNOWN_ERROR,	TEXT(T_SOCKS_UNKNOWN_ERROR)},
+	{S_SSL_ERROR,		TEXT_(T_SSL_ERROR)},
+	{S_NO_SSL,		TEXT_(T_NO_SSL)},
+	{S_BAD_SOCKS_VERSION,	TEXT_(T_BAD_SOCKS_VERSION)},
+	{S_SOCKS_REJECTED,	TEXT_(T_SOCKS_REJECTED_OR_FAILED)},
+	{S_SOCKS_NO_IDENTD,	TEXT_(T_SOCKS_NO_IDENTD)},
+	{S_SOCKS_BAD_USERID,	TEXT_(T_SOCKS_BAD_USERID)},
+	{S_SOCKS_UNKNOWN_ERROR,	TEXT_(T_SOCKS_UNKNOWN_ERROR)},
 
-	{S_BLOCKED_URL,		TEXT(T_BLOCKED_URL)},
-	{S_NO_PROXY,		TEXT(T_NO_PROXY)},
-	{S_NO_SMB_CLIENT,	TEXT(T_NO_SMB_CLIENT)},
+	{S_BLOCKED_URL,		TEXT_(T_BLOCKED_URL)},
+	{S_NO_PROXY,		TEXT_(T_NO_PROXY)},
+	{S_NO_SMB_CLIENT,	TEXT_(T_NO_SMB_CLIENT)},
+
+	{S_WAIT_REDIR,		TEXT_(T_WAITING_FOR_REDIRECT_CONFIRMATION)},
 	{0,			NULL}
 };
