@@ -46,29 +46,48 @@ static struct list_head timers = {&timers, &timers};
 ttime get_time(void)
 {
 	struct timeval tv;
-	gettimeofday(&tv, NULL);
-	return tv.tv_sec * 1000 + tv.tv_usec / 1000;
+	int rs;
+	EINTRLOOP(rs, gettimeofday(&tv, NULL));
+	return (uttime)tv.tv_sec * 1000 + tv.tv_usec / 1000;
 }
 
 int can_write(int fd)
 {
 	fd_set fds;
 	struct timeval tv = {0, 0};
+	int rs;
 	FD_ZERO(&fds);
+	if (fd < 0)
+		internal("can_write: handle %d", fd);
+	if (fd >= (int)FD_SETSIZE) {
+		error("too big handle %d", fd);
+		fatal_tty_exit();
+		exit(RET_FATAL);
+	}
 	FD_SET(fd, &fds);
-	return select(fd + 1, NULL, &fds, NULL, &tv);
+	EINTRLOOP(rs, select(fd + 1, NULL, &fds, NULL, &tv));
+	return rs;
 }
 
 int can_read(int fd)
 {
 	fd_set fds;
 	struct timeval tv = {0, 0};
+	int rs;
 	FD_ZERO(&fds);
+	if (fd < 0)
+		internal("can_read: handle %d", fd);
+	if (fd >= (int)FD_SETSIZE) {
+		error("too big handle %d", fd);
+		fatal_tty_exit();
+		exit(RET_FATAL);
+	}
 	FD_SET(fd, &fds);
-	return select(fd + 1, &fds, NULL, NULL, &tv);
+	EINTRLOOP(rs, select(fd + 1, &fds, NULL, NULL, &tv));
+	return rs;
 }
 
-long select_info(int type)
+unsigned long select_info(int type)
 {
 	int i = 0, j;
 	struct cache_entry *ce;
@@ -81,7 +100,7 @@ long select_info(int type)
 			foreach(ce, timers) i++;
 			return i;
 		default:
-			internal("cache_info: bad request");
+			internal("select_info_info: bad request");
 	}
 	return 0;
 }
@@ -93,7 +112,7 @@ struct bottom_half {
 	void *data;
 };
 
-struct list_head bottom_halves = { &bottom_halves, &bottom_halves };
+static struct list_head bottom_halves = { &bottom_halves, &bottom_halves };
 
 int register_bottom_half(void (*fn)(void *), void *data)
 {
@@ -145,18 +164,11 @@ static ttime last_time;
 
 static void check_timers(void)
 {
-	ttime interval = get_time() - last_time;
-	struct timer *t;
-	foreach(t, timers) t->interval -= interval;
+	ttime interval = (uttime)get_time() - (uttime)last_time;
+	struct timer * volatile t;	/* volatile because of setjmp */
+	foreach(t, timers) t->interval -= (uttime)interval;
 	/*ch:*/
 	foreach(t, timers) if (t->interval <= 0) {
-		/*struct timer *tt = t;
-		del_from_list(tt);
-		tt->func(tt->data);
-		mem_free(tt);
-		CHK_BH;
-		goto ch;*/
-
 		struct timer *tt;
 #ifdef DEBUG_CALLS
 		fprintf(stderr, "call: timer %p\n", t->func);
@@ -174,7 +186,7 @@ static void check_timers(void)
 		mem_free(t);
 		t = tt;
 	} else break;
-	last_time += interval;
+	last_time += (uttime)interval;
 }
 
 int install_timer(ttime t, void (*func)(void *), void *data)
@@ -211,8 +223,12 @@ void kill_timer(int id)
 
 void *get_handler(int fd, int tp)
 {
-	if (fd < 0 || fd >= (int)FD_SETSIZE) {
-		internal("get_handler: handle %d >= FD_SETSIZE %d", fd, FD_SETSIZE);
+	if (fd < 0)
+		internal("get_handler: handle %d", fd);
+	if (fd >= (int)FD_SETSIZE) {
+		error("too big handle %d", fd);
+		fatal_tty_exit();
+		exit(RET_FATAL);
 		return NULL;
 	}
 	switch (tp) {
@@ -227,8 +243,12 @@ void *get_handler(int fd, int tp)
 
 void set_handlers(int fd, void (*read_func)(void *), void (*write_func)(void *), void (*error_func)(void *), void *data)
 {
-	if (fd < 0 || fd >= (int)FD_SETSIZE) {
-		internal("set_handlers: handle %d >= FD_SETSIZE %d", fd, FD_SETSIZE);
+	if (fd < 0)
+		internal("set_handlers: handle %d", fd);
+	if (fd >= (int)FD_SETSIZE) {
+		error("too big handle %d", fd);
+		fatal_tty_exit();
+		exit(RET_FATAL);
 		return;
 	}
 	threads[fd].read_func = read_func;
@@ -276,15 +296,17 @@ static int signal_pipe[2];
 
 static void signal_break(void *data)
 {
-	char c;
-	while (can_read(signal_pipe[0])) read(signal_pipe[0], &c, 1);
+	unsigned char c;
+	while (can_read(signal_pipe[0])) {
+		int rd;
+		EINTRLOOP(rd, read(signal_pipe[0], &c, 1));
+		if (rd != 1) break;
+	}
 }
 
-static void got_signal(int sig)
+SIGNAL_HANDLER static void got_signal(int sig)
 {
-#ifndef BEOS
 	int sv_errno = errno;
-#endif
 		/*fprintf(stderr, "ERROR: signal number: %d\n", sig);*/
 	if (sig >= NUM_SIGNALS || sig < 0) {
 		/*error("ERROR: bad signal number: %d", sig);*/
@@ -297,44 +319,49 @@ static void got_signal(int sig)
 	}
 	signal_mask[sig] = 1;
 	ret:
-	if (can_write(signal_pipe[1])) write(signal_pipe[1], "x", 1);
-#ifndef BEOS
+	if (can_write(signal_pipe[1])) {
+		int wr;
+		EINTRLOOP(wr, write(signal_pipe[1], "x", 1));
+	}
 	errno = sv_errno;
-#endif
 }
 
 static struct sigaction sa_zero;
 
 void install_signal_handler(int sig, void (*fn)(void *), void *data, int critical)
 {
+	int rs;
 	struct sigaction sa = sa_zero;
 	if (sig >= NUM_SIGNALS || sig < 0) {
 		internal("bad signal number: %d", sig);
 		return;
 	}
 	if (!fn) sa.sa_handler = SIG_IGN;
-	else sa.sa_handler = got_signal;
+	else sa.sa_handler = (void *)got_signal;
 	sigfillset(&sa.sa_mask);
 	sa.sa_flags = SA_RESTART;
-	if (!fn) sigaction(sig, &sa, NULL);
+	if (!fn)
+		EINTRLOOP(rs, sigaction(sig, &sa, NULL));
 	signal_handlers[sig].fn = fn;
 	signal_handlers[sig].data = data;
 	signal_handlers[sig].critical = critical;
-	if (fn) sigaction(sig, &sa, NULL);
+	if (fn)
+		EINTRLOOP(rs, sigaction(sig, &sa, NULL));
 }
 
 void interruptible_signal(int sig, int in)
 {
 	struct sigaction sa = sa_zero;
+	int rs;
 	if (sig >= NUM_SIGNALS || sig < 0) {
 		internal("bad signal number: %d", sig);
 		return;
 	}
 	if (!signal_handlers[sig].fn) return;
-	sa.sa_handler = got_signal;
+	sa.sa_handler = (void *)got_signal;
 	sigfillset(&sa.sa_mask);
 	if (!in) sa.sa_flags = SA_RESTART;
-	sigaction(sig, &sa, NULL);
+	EINTRLOOP(rs, sigaction(sig, &sa, NULL));
 }
 
 static int check_signals(void)
@@ -360,10 +387,13 @@ static int check_signals(void)
 
 static void sigchld(void *p)
 {
+	pid_t pid;
 #ifndef WNOHANG
-	wait(NULL);
+	EINTRLOOP(pid, wait(NULL));
 #else
-	while (waitpid(-1, NULL, WNOHANG) > 0) ;
+	do {
+		EINTRLOOP(pid, waitpid(-1, NULL, WNOHANG));
+	} while (pid > 0);
 #endif
 }
 
@@ -377,7 +407,10 @@ int terminate_loop = 0;
 void select_loop(void (*init)(void))
 {
 	struct stat st;
-	if (stat(".", &st) && getenv("HOME")) chdir(getenv("HOME"));
+	int rs;
+	EINTRLOOP(rs, stat(".", &st));
+	if (rs && getenv("HOME"))
+		EINTRLOOP(rs, chdir(getenv("HOME")));
 	memset(&sa_zero, 0, sizeof sa_zero);
 	memset(signal_mask, 0, sizeof signal_mask);
 	memset(signal_handlers, 0, sizeof signal_handlers);
@@ -389,23 +422,22 @@ void select_loop(void (*init)(void))
 	ignore_signals();
 	if (c_pipe(signal_pipe)) {
 		error("ERROR: can't create pipe for signal handling");
-		retval = RET_FATAL;
-		return;
+		fatal_tty_exit();
+		exit(RET_FATAL);
 	}
-	fcntl(signal_pipe[0], F_SETFL, O_NONBLOCK);
-	fcntl(signal_pipe[1], F_SETFL, O_NONBLOCK);
+	EINTRLOOP(rs, fcntl(signal_pipe[0], F_SETFL, O_NONBLOCK));
+	EINTRLOOP(rs, fcntl(signal_pipe[1], F_SETFL, O_NONBLOCK));
 	set_handlers(signal_pipe[0], signal_break, NULL, NULL, NULL);
 	init();
 	CHK_BH;
 	while (!terminate_loop) {
-		int n, i;
+		volatile int n;	/* volatile because of setjmp */
+		int i;
 		struct timeval tv;
 		struct timeval *tm = NULL;
 		check_signals();
 		check_timers();
 		check_timers();
-		check_timers();
-		check_timers();
 		if (!F) redraw_all_terminals();
 #ifdef OS_BAD_SIGNALS
 	/* Cygwin has buggy signals that sometimes don't interrupt select.
@@ -415,7 +447,7 @@ void select_loop(void (*init)(void))
 		tm = &tv;
 #endif
 		if (!list_empty(timers)) {
-			ttime tt = ((struct timer *)timers.next)->interval + 1;
+			ttime tt = (uttime)((struct timer *)timers.next)->interval + 1;
 			if (tt < 0) tt = 0;
 #ifdef OS_BAD_SIGNALS
 			if (tt < 1000)
@@ -455,7 +487,11 @@ void select_loop(void (*init)(void))
 #ifdef DEBUG_CALLS
 			fprintf(stderr, "select intr\n");
 #endif
-			if (errno != EINTR) error("ERROR: select failed: %d", errno);
+			if (errno != EINTR) {
+				error("ERROR: select failed: %s", strerror(errno));
+				fatal_tty_exit();
+				exit(RET_FATAL);
+			}
 			continue;
 		}
 #ifdef DEBUG_CALLS
